import os
from multiprocessing import cpu_count


def get_cpu_cores_limit():
    cpu_cores = cpu_count()
    cores_reserve_for_host = 1
    if not cpu_cores or cpu_cores <= cores_reserve_for_host:
        result = 1
    else:
        result = cpu_cores - cores_reserve_for_host
    return result


def get_system_memory_size():
    return os.sysconf("SC_PAGE_SIZE") * os.sysconf("SC_PHYS_PAGES")


def get_memory_size_limit():
    total_mem_bytes = get_system_memory_size()
    bytes_reserve_for_host = int(2 * 1024.0**3)  # Reserve 2 Gb by default
    if total_mem_bytes <= bytes_reserve_for_host:
        result = int(total_mem_bytes * 0.7)
    else:
        result = int(total_mem_bytes - bytes_reserve_for_host)
    return result


def get_default_shm_size():
    total_mem_bytes = get_system_memory_size()
    return int(total_mem_bytes / 6)



from pathlib import Path

import docker.errors as docker_errors
from textual import on, work
from textual.containers import Horizontal, ScrollableContainer
from textual.screen import ModalScreen
from textual.widgets import (
    Button,
    Collapsible,
    DataTable,
    Footer,
    Label,
    Log,
    Static,
)

from .server_cfg import BaseFunctions
from .utils.infra import get_instanse_token, get_sinara_servers
from .utils.process import decode_lines


class RunningScreen(ModalScreen, BaseFunctions):
    CSS_PATH = "style.css"

    BINDINGS = [
        ("escape", "app.pop_screen", "Back to main."),
        ("ctrl+s", "save_screen", "Save screenshot"),
        ("f6", "copy_logs", "Copy log to clipboard"),
        ("u", ""),
        ("s", ""),
    ]

    def action_save_screen(self):
        image_folder = Path("./images/")
        image_folder.mkdir(exist_ok=True)
        self.app.save_screenshot(image_folder.joinpath("running.svg"))

    def compose(self):
        self.selected_uid = None
        yield Footer()
        with ScrollableContainer():
            yield Label("Running servers:")
            yield Static("Click on server for inspect server url.")
            yield DataTable(cursor_type="row")

            yield Static()

            with Collapsible(title="not supported by sinaraml", collapsed=False):
                yield Static("Select server in click!")
                self.sudo_button = Button(
                    "install sudo",
                    id="install_sudo",
                    classes="button",
                    disabled=True,
                )
                yield self.sudo_button
                self.data_button = Button(
                    "remove data & tmp",
                    id="remove_data_button",
                    classes="button",
                    disabled=True,
                )
                yield self.data_button

            yield Button(
                "Refresh servers",
                id="refresh_button",
                classes="button",
                variant="success",
            )

            with Horizontal():
                yield Button(
                    "< Back",
                    id="back_button",
                    classes="back_button button",
                    variant="primary",
                )
                yield Button(
                    "Exit",
                    id="exit_button",
                    classes="exit_button button",
                    variant="error",
                )

            self.log_window: Log = Log(
                highlight=True,
                id="running_screen_log_window",
                classes="log_window",
            )
            yield self.log_window

    @work(thread=True)
    def mount_servers(self):
        servers = get_sinara_servers()
        table = self.query_one(DataTable)
        table.clear()
        table.columns.clear()

        if len(servers[0]) > 0:
            table.add_columns(*servers[0])
            table.add_rows(servers[1:])

    def on_mount(self) -> None:
        self.mount_servers()

    @on(Button.Pressed, "#refresh_button")
    def refresh_button(self):
        self.mount_servers()

    @work(thread=True)
    def select_server(self, row):
        instanceName = row[1]
        port = row[2]
        try:
            server_url = get_instanse_token(instanceName, port)
            out_string = f"{instanceName} running on: {server_url}\n"
            self.log_window.write_line(out_string)
            self.log_window.write_line("-----")
        except docker_errors.APIError:
            self.log_window.write_line("Server not running")
            self.log_window.write_line("-----")

    @on(DataTable.RowSelected)
    def on_row_selected(self, event: DataTable.RowSelected):
        row = event.data_table.get_row_at(event.cursor_row)
        self.select_server(row)
        self.sudo_button.disabled = False
        self.data_button.disabled = False
        self.selected_uid = row[0]

    @work(thread=True)
    def install_sudo(self, uid):
        exec_cmd = "docker exec -it -u root {uid} bash -c '{cmd}'"
        cmds = [
            "rm -rf /etc/apt/sources.list.d/cuda*",
            "chmod 777 /tmp",
            "apt-get update",
            "apt-get install sudo -y",
            "echo ${NB_USER} ALL=NOPASSWD: ALL >> /etc/sudoers.d/${NB_USER}",
        ]
        self.log_window.loading = True
        stop = False
        for sub_cmd in cmds:
            for lines in decode_lines(exec_cmd.format(cmd=sub_cmd, uid=uid)):
                if len(lines) > 0:
                    self.write_log_lines(lines)
                    if "failed" in lines[-1]:
                        stop = True
            if stop:
                break

        self.log_window.loading = False
        if not stop:
            self.write_log_lines([f"Hack sudo for {uid} done [True]!"])

    @on(Button.Pressed, "#install_sudo")
    def install_sudo_button(self):
        if self.selected_uid is not None:
            self.install_sudo(self.selected_uid)

    @work(thread=True)
    def _remove_data(self, uid):
        exec_cmd = "docker exec -it -u root {uid} bash -c '{cmd}'"
        cmds = [
            "rm -rf /data/*",
            "rm -rf /tmp/*",
        ]
        self.log_window.loading = True
        stop = False
        for sub_cmd in cmds:
            for lines in decode_lines(exec_cmd.format(cmd=sub_cmd, uid=uid)):
                if len(lines) > 0:
                    self.write_log_lines(lines)
                    if "failed" in lines[-1]:
                        stop = True
            if stop:
                break

        self.log_window.loading = False
        if not stop:
            self.write_log_lines([f"/data/* and /tmp/* removed for {uid} done [True]!"])

    @on(Button.Pressed, "#remove_data_button")
    def remove_data_button(self):
        if self.selected_uid is not None:
            self._remove_data(self.selected_uid)

    @on(Button.Pressed, "#back_button")
    def back_button(self):
        self.log_window.clear()
        self.selected_uid = None
        self.dismiss()

    @on(Button.Pressed, "#exit_button")
    def exit_button(self):
        self.parent.exit()



import json
import os
from pathlib import Path

from textual import work

from .utils import (
    FilteredConfigTree,
    decode_lines,
    generate_from_screen,
    load_from_file,
)


class BaseFunctions:
    def action_copy_logs(self):
        self.app.copy_to_clipboard("\n".join(self.log_window.lines))

    def write_log_lines(self, lines):
        self.log_window.clear()

        if type(lines) is list:
            self.log_window.write_lines(lines)

        elif type(lines) is str:
            self.log_window.write(lines)

        else:
            self.log_window.write(f"{type(lines)=} not supported!")

        self.app.refresh(layout=True)

    @work(thread=True)
    def cmd(self, cmd: str):
        cmd = cmd.replace("'store_true'", "")
        self.write_log_lines([cmd + "\n"])
        for lines in decode_lines(cmd):
            self.write_log_lines([cmd + "\n"] + lines[1:])


class ServerFunctions(BaseFunctions):
    def generate_config(self):
        result = generate_from_screen(self)
        return result

    @work(thread=True)
    def reload_config_dir(self):
        sinaraX_path = Path().home().joinpath(".sinaraX")
        sinaraX_path.mkdir(exist_ok=True)
        self.config_tree.path = sinaraX_path

    @work(thread=True)
    def save_cfg(self):
        result = self.generate_config()
        if result:
            sinaraX_path = Path().home().joinpath(".sinaraX")
            sinaraX_path.mkdir(exist_ok=True)
            config_name = sinaraX_path.joinpath(
                self.config_dict.get("instanceName", "base.") + ".json"
            )
            with open(config_name, "w") as fd:
                json.dump(self.config_dict, fd)

            self.log_window.write_line(f"Config {config_name} saved!")
            self.reload_config_dir()
        else:
            self.log_window.write_line("Config not saved!")

    def select_config(self, event: FilteredConfigTree.FileSelected):
        self.selected_config_path = event.path.as_posix()

    def remove_config_file_button(self):
        selected_file = Path(str(self.selected_config_path))
        if selected_file.is_file():
            os.remove(selected_file.as_posix())
            self.log_window.write_line(f"Config {selected_file} removed!")

        self.reload_config_dir()

    def load_cfg_button(self):
        selected_file = Path(str(self.selected_config_path))
        if selected_file.is_file():
            load_from_file(self, selected_file.as_posix())
            self.log_window.write_line(f"[True] Config {selected_file} loaded!")
        else:
            self.log_window.write_line("[False] Config not selected!")
            self.notify("Config not selected!", severity="error")

        self.reload_config_dir()

    def get_cfg_button(self):
        self.log_window.clear()
        result = self.generate_config()
        if result:
            self.write_log_lines(json.dumps(self.config_dict, indent=4))

    def save_cfg_button(self):
        self.save_cfg()



import getpass
import importlib.metadata
import json
import platform
import re

import requests

from .process import start_cmd


def update_sinara_org():
    """Update sinara org."""
    cmd = "sinara org update"
    failed = False
    lines = []
    for line in start_cmd(cmd):
        lines.append(line)
        if "failed" in line:
            failed = True

        if "docker: not found" in line:
            failed = True
    return lines, failed


def check_docker():
    """Check docker."""
    check_docker_cmd = "docker info -f json"
    lines = []
    failed = False
    for line in start_cmd(check_docker_cmd):
        lines.append(line)
        if "failed" in line:
            failed = True

        if "docker: not found" in line:
            failed = True

    result = {
        "ok": not failed,
        "ServerVersion": None,
        "NCPU": None,
        "MemTotal": None,
        "ServerErrors": [],
    }

    if not failed:
        output = "".join(lines)
        start_json = output.find("{")
        end_json = len(output) - output[::-1].find("}")
        if (start_json != -1) and (end_json != -1):
            docker_info = json.loads(output[start_json:end_json])
            result["NCPU"] = docker_info.get("NCPU")
            result["ServerErrors"] = docker_info.get("ServerErrors", [])
            result["ServerVersion"] = docker_info.get("ServerVersion")
            result["MemTotal"] = round(
                int(docker_info.get("MemTotal")) / 1024 / 1024, 2
            )
            for err in result.get("ServerErrors"):
                if "error during connect" in err:
                    result["ok"] = False
    return result


def check_platform():
    """Check platform."""
    _platform = platform.system()
    is_windows = "Windows" in _platform
    result = {
        "ok": not is_windows,
        "platform": _platform,
    }
    return result


def check_docker_group():
    """Check docker group."""
    result = {
        "docker_group": False,
        "user_in_group": False,
        "is_unix": False,
        "username": getpass.getuser(),
    }

    try:
        import grp

        docker_group = grp.getgrnam("docker")
        result["docker_group"] = True
        result["is_unix"] = True
        result["user_in_group"] = (result["username"] in docker_group.gr_mem) or (
            result["username"] == "root"
        )
    except ModuleNotFoundError:
        pass
    except KeyError:
        pass

    return result


def get_sinara_servers():
    """Get sinara servers."""
    docker_stats = check_docker()
    containers: list[dict] = []
    if docker_stats["ok"]:
        cmd = "docker ps -a --filter 'label=sinaraml.platform' --format json"

        for line in start_cmd(cmd):
            try:
                decoded_line = json.loads(line)
            except json.decoder.JSONDecodeError:
                continue

            ports = decoded_line.get("Ports", "-,-").split(",")
            ports = [
                port.replace("->8888/tcp", "").split(":")[1]
                for port in ports
                if "->8888/tcp" in port
            ]

            ports = [port for port in ports if len(port.strip()) > 0]

            if len(ports) == 0:
                port = 8888
            else:
                port = int(ports[-1])

            image = decoded_line.get("Image", "no image")
            containers.append(
                {
                    "id": decoded_line.get("ID"),
                    "instanceName": decoded_line.get("Names", "no name"),
                    "port": port,
                    "image": "cv" if "cv" in image else "ml",
                    "exp": "exp" in image,
                }
            )

    label_row = ["id", "instanceName", "port", "image", "exp"]
    rows = []
    for row in containers:
        for index in list(row):
            if index not in label_row:
                label_row.append(index)
        rows.append(list(row.values()))

    return [label_row] + rows


def get_server_url(instance):
    url = None
    commands = [
        "jupyter lab list",
        "jupyter server list",
        "jupyter notebook list",
    ]
    for cmd in commands:
        if url:
            continue

        for line in start_cmd(f"docker exec {instance} {cmd}"):
            if url:
                continue

            if any(x in line for x in ["http://", "https://"]):
                m = re.search(r"(http[^\s]+)", line)
                url = m.group(1) if m else None

    return url


def get_server_token(server_url):
    m = re.search(r"token=([a-f0-9-][^\s]+)", server_url)
    return m.group(1) if m else None


def get_server_protocol(server_url):
    m = re.search(r"^(http:|https:)", server_url)
    return str(m.group(1))[:-1] if m else None


def get_instanse_token(instanceName, host_port):
    url = get_server_url(instanceName)

    try:
        token = get_server_token(url)
    except TypeError:
        token = None

    token_str = f"?token={token}" if token else ""

    try:
        protocol = get_server_protocol(url)
    except TypeError:
        protocol = "http"

    server_ip = "localhost"
    server_url = f"{protocol}://{server_ip}:{host_port}/{token_str}"
    return server_url


def check_last_version(name: str):
    try:
        local_version = str(importlib.metadata.version(name)).lower()

        url = f"https://pypi.org/pypi/{name}/json"
        resp = requests.get(url)
        if resp.status_code == 200:
            package_data = resp.json()

            last_version = (
                package_data.get("info", {}).get("version", "not found!").lower()
            )

            if last_version != local_version:
                return (
                    False,
                    f"Available new! [{local_version}] < [{last_version}]",
                )
        else:
            return False, "Pypi not available from this env!"
    except UnboundLocalError:
        return False, "UnboundLocalError!"
    except ImportError:
        return False, "ImportError!"

    return True, "Latest!"


if __name__ == "__main__":
    get_sinara_servers()



import enum
import json
from pathlib import Path
from typing import Iterable

from textual.widgets import DirectoryTree, RadioSet

from .sinara_server_utils import get_memory_size_limit


class SinaraImageType(enum.Enum):
    CV = 0
    ML = 1


class SinaraRunMode(enum.Enum):
    Quick = 0
    Basic = 1


class FilteredConfigTree(DirectoryTree):
    def filter_paths(self, paths: Iterable[Path]) -> Iterable[Path]:
        return [
            path for path in paths if (path.is_file() and path.name.endswith(".json"))
        ]


def remap_radio_set(_dict: dict, key: str, remap_dict: dict):
    if _dict.get(key) is not None:
        _dict[key] = remap_dict.get(_dict[key])


def remap_checkbox(_dict: dict, key: str, remap_dict: dict):
    if _dict.get(key) is not None:
        _dict[key] = remap_dict.get(_dict[key], False)


def remap_config(_dict: dict, _from_screen=True):
    _remap_radio = {
        "runMode": {
            SinaraRunMode.Quick.value: "q",
            SinaraRunMode.Basic.value: "b",
        },
        "sinara_image_num": {
            SinaraImageType.CV.value: "2",
            SinaraImageType.ML.value: "1",
        },
    }

    _remap_checkbox = {
        "gpuEnabled": {True: "y", False: "n"},
        "experimental": {True: "store_true", False: None},
        "insecure": {True: "store_true", False: None},
    }

    if _dict.get("memLimit"):
        if _from_screen:
            _dict["memLimit"] = int(float(_dict.get("memLimit")) * 1024 * 1024 * 1024)
        else:
            memLimit = str(_dict["memLimit"])
            if "g" in memLimit:
                memLimit = int(memLimit.replace("g", ""))
            elif "m" in memLimit:
                memLimit = float(memLimit.replace("m", "")) / 1024
            elif type(_dict["memLimit"]) is int:
                max_mem = get_memory_size_limit()
                if int(memLimit) > max_mem:
                    memLimit = max_mem

                memLimit = int(memLimit) // 1024 // 1024 // 1024
            else:
                memLimit = None
                del _dict["memLimit"]

            if memLimit is not None:
                _dict["memLimit"] = str(memLimit)

    if _dict.get("cpuLimit"):
        if _from_screen:
            _dict["cpuLimit"] = int(_dict["cpuLimit"])
        else:
            _dict["cpuLimit"] = str(_dict["cpuLimit"])

    for key, val in _remap_radio.items():
        remap_radio_set(
            _dict, key, val if _from_screen else {v: k for k, v in val.items()}
        )

    for key, val in _remap_checkbox.items():
        remap_checkbox(
            _dict, key, val if _from_screen else {v: k for k, v in val.items()}
        )

    sinara_images = [
        ["buslovaev/sinara-notebook", "buslovaev/sinara-cv"],
        ["buslovaev/sinara-notebook-exp", "buslovaev/sinara-cv-exp"],
    ]
    if _from_screen:
        image_type_id = int(_dict.get("experimental") is not None)
        image = sinara_images[image_type_id][int(_dict.get("sinara_image_num")) - 1]
        _dict["image"] = image
        _dict["serverType"] = "cv" if "cv" in image else "ml"
        del _dict["sinara_image_num"]
    else:
        image = _dict.get("image")
        if image:
            if "exp" in image:
                _dict["experimental"] = True

            if "cv" in image:
                _dict["sinara_image_num"] = SinaraImageType.CV.value
            else:
                _dict["sinara_image_num"] = SinaraImageType.ML.value


def load_from_file(screen, file):
    with open(file) as fd:
        config = json.load(fd)

    remap_config(config, False)

    for child in screen.walk_children():
        if config.get(child.name) is not None:
            if type(child) is RadioSet:
                child._selected = config[child.name]
                child.action_toggle_button()
            else:
                child.value = config[child.name]


def generate_from_screen(screen):
    for child in screen.walk_children():
        if child.name is not None:
            if type(child) is RadioSet:
                screen.config_dict[child.name] = child._selected
            else:
                screen.config_dict[child.name] = child.value

    remap_config(screen.config_dict)

    try:
        if screen.config_dict.get("runMode", "") == "b":
            for _key in ["jovyanRootPath"]:
                assert len(screen.config_dict.get(_key, "")) > 0, f"{_key} empty!"
    except AssertionError as e:
        screen.write_log_lines("No path specified for:\n" + str(e))
        return False

    screen.config_dict = {
        key: val
        for key, val in screen.config_dict.items()
        if (val is not None) and len(str(val)) > 0
    }

    return True
